@@ -1,11 +1,13 @@
 from bisect import bisect
-from typing import Tuple
+from typing import Tuple, Union, List
 
 import numpy as np
 import pandas as pd
 
 from .configs import NodeDataConfig
 from .loader import load_y
+import logging
+logger = logging.getLogger(__name__)
 
 
 class DataPrep:
@@ -63,32 +65,78 @@ class BinFeaturesDataPrep(DataPrep):
         Number of bins to use
     """
 
-    def __init__(self, data: NodeDataConfig, nbins):
-        super().__init__(data, nbins=nbins)
+    def __init__(self, data: NodeDataConfig, nbins: int, bins: dict):
+        super().__init__(data, nbins=nbins, bins=bins)
 
     def dataprep_method(self, x, y):
 
-        def find_bins(xx, nbins: int):
+        def find_bins(xx, nbins: int) -> np.ndarray:
+            """
+            Function used to find the bins to discretize xcol in nbins modalities
+
+            Parameters
+            ----------
+            xx : pd.Series
+               Series to discretize
+
+            nbins: int
+                number of modalities
+
+            Return
+            ------
+            bins: np.ndarray
+               the bins for disretization (result from numpy percentile function)
+            """
             q_list = np.arange(100.0 / nbins, 100.0, 100.0 / nbins)
             bins = np.array([np.nanpercentile(xx, i) for i in q_list])
             return bins
 
-        def get_bins(xx, nb_bucket: int):
+        def get_bins(xx: pd.Series, nb_bucket: int) -> np.ndarray:
+            logger.debug(f"Getting bins for {xx.name}...")
+            if nb_bucket == 0:
+                raise ValueError("nb_bucket must be greater than 0")
+            if nb_bucket == 1:
+                return np.array([])
+            if len(xx) == 0 or len(xx) == 1:
+                return np.array([])
+            if len(np.unique(xx)) <= nb_bucket:
+                return np.ediff1d(np.unique(xx)) / 2. + np.unique(xx)[:-1]
             bins = find_bins(xx, nb_bucket)
             while len(set(bins.round(5))) != len(bins):
                 nb_bucket -= 1
                 bins = find_bins(xx, nb_bucket)
             if len(bins) != nb_bucket - 1:
                 raise ValueError(f"Error in get_bins : {len(bins) + 1} bins where found but {nb_bucket} were asked.")
+            logger.debug(f"... got bins for {xx.name}")
             return bins
 
-        def dicretize(x_series):
-            # noinspection PyUnresolvedReferences
-            bins = get_bins(x_series, self.nbins)
-            mask = np.isnan(x_series)
-            discrete_x = x_series.apply(lambda var: bisect(bins, var))
+        def discretize(xx: pd.Series, bins: Union[np.ndarray, List[float]]) -> pd.Series:
+            """
+            Transform a Series of float to a Series if int, where each float is replaced by the bin it matches.
+
+            Parameters
+            ----------
+            xx : pd.Series
+                Series to discretize
+            bins : Union[np.ndarray, List[float]]
+                The list of bins in the form e.g. [0.1, 0.3, ... 0.8] :
+                    // -inf ---> 0.1 is bin 0
+                    // 0.1 ---> 0.3 is bin 1, etc.
+                    // 0.8 ---> +inf is the last bin
+
+            Return
+            ------
+            discrete_x : pd.Series
+                The discretization of x
+
+            """
+            logger.debug(f"Discretizing {xx.name}...")
+            mask = np.isnan(xx)
+            discrete_x = xx.apply(lambda var: bisect(bins, var))
             discrete_x[mask] = np.nan
+            logger.debug(f"... discretized {xx.name}")
             return discrete_x
 
-        x = x.apply(lambda xx: dicretize(xx), axis=0)
+        # noinspection PyUnresolvedReferences
+        x = x.apply(lambda xx: discretize(xx, self.bins.get(xx.name, get_bins(xx, self.nbins))), axis=0)
         return x, y
